package scala.tools.nsc
package backend
package icode

import scala.tools.nsc.plugins
import java.io._
import plugins.Plugin
import tools.nsc._

/** The actual work is done here. */
abstract class GenCSharp extends Plugin
{
  /** The compiler.*/
  val global: Global

  import global._
  import icodes._

  def generateOutput()

  class GenCSharpPhase(prev: Phase) extends ICodePhase(prev)
  {
    def name = GenCSharpPlugin.PluginName

    override def run {
      if (settings.debug.value) inform("[running phase " + name + " on icode]")
      if (settings.Xdce.value)
        for ((sym, cls) <- icodes.classes ; if inliner.isClosureClass(sym) && !deadCode.liveClosures(sym))
          icodes.classes -= sym

      icodes.classes.values foreach apply

      generateOutput()
    }

    val tfa = new analysis.MethodTFA
    // TODO confirm: the TypeKind on top of the stack is the last in the List
    val preTypeStack = scala.collection.mutable.Map[(IMethod, BasicBlock, Int), List[TypeKind]]()
    val postTypeStack = scala.collection.mutable.Map[(IMethod, BasicBlock, Int), List[TypeKind]]()

    def populateTypeStacks(method: IMethod) {
      tfa.init(method)
      tfa.run
      for (currentBlock <- method.code.blocks) {
        var tfainfo = tfa.in(currentBlock)
        for ((ins, idx) <- currentBlock.toList.zipWithIndex) {
          val tfa.lattice.IState(entryBindings, typeStack: icodes.TypeStack) = tfainfo
          var listOfTypeKinds : List[TypeKind] = typeStack.types
          preTypeStack += ((method, currentBlock, idx) -> listOfTypeKinds)
          tfainfo = tfa.interpret(tfainfo, ins)
          val tfa.lattice.IState(entryBindings2, typeStack2) = tfainfo
          postTypeStack += ((method, currentBlock, idx) -> typeStack2.types)
        }
      }
    }

    val rdef = new icodes.reachingDefinitions.ReachingDefinitionsAnalysis
    case class StackSlot(height: Int, mm: IMethod, bb: BasicBlock, idx: Int)
    val depends = scala.collection.mutable.Map[(IMethod, BasicBlock, Int), Set[StackSlot]]()
    val emptySetOfStackSlots = Set[StackSlot]()

    def populateDependsMap(method: IMethod) {
      scala.Console.println("method: " + method.toString)
      rdef.init(method)
      rdef.run
      for (currentBlock <- method.code.blocks) {
        var rdefinfo = rdef.in(currentBlock)
        for ((currentIns, currentIdx) <- currentBlock.toList.zipWithIndex) {
          val icodes.reachingDefinitions.rdefLattice.IState(rdefVars,
                                                            rdefStack : List[Set[(BasicBlock, Int)]]) = rdefinfo
          makeSenseOfReachingDefs(rdefVars, rdefStack)
          /** Return the reaching definitions corresponding to the point after idx. */
          rdefinfo = rdef.interpret(currentBlock, currentIdx, rdefinfo)

          def makeSenseOfReachingDefs(rdefVars:   Set[(Local, BasicBlock, Int)],
                                      rdefStack : List[Set[(BasicBlock, Int)]]) {
            for((slot, depth) <- rdefStack.zipWithIndex; (bb, bbidx) <- slot){
              val msg = where(bb, bbidx) + " pushed a value that reaches " + where(currentBlock, currentIdx) + " with depth " + depth 
              scala.Console.println(msg)
            }
            def where(b: BasicBlock, i: Int) = "/" + b(i) + " in block " + b + "\\"
          }

        }
        // TODO the rdefinfo obtained in the last iteration need not be added, right? Because at the beginning of the next basic block(s) it is added anyway
      }


    }

    def makeItLookGoodOnTheConsole(method: IMethod) {
      scala.Console.println("method: " + method.toString)
      for (currentBlock <- method.code.blocks) {
        scala.Console.println("\t block: " + currentBlock)
        for ((ins, idx) <- currentBlock.toList.zipWithIndex) {
          scala.Console.println("\t\t\t  type stack : " + preTypeStack(method, currentBlock, idx).mkString("[", ",", "]"))
          scala.Console.println("\t\t" + idx + "|\t" + ins)
        }
        scala.Console.println("\t\t\t  last type stack : " + postTypeStack(method, currentBlock, currentBlock.length-1).mkString("[", ",", "]"))
      }
    }

    override def apply(cls: IClass) {
      for (method <- cls.methods if method.code != null) {
        populateTypeStacks(method)
        populateDependsMap(method)
      }
      for (method <- cls.methods if method.code != null) {
        makeItLookGoodOnTheConsole(method)
      }
    }

  }


  class CSharpTextPrinter(writer: PrintWriter, lin: Linearizer) extends TextPrinter(writer, lin) {

    override def printBlock(bb: BasicBlock) {
      print(bb.label)
      if (bb.loopHeader) print("[loop header]")
      print(": ");
      if (settings.debug.value) print("pred: " + bb.predecessors + " succs: " + bb.successors + " flags: " + bb.flagsString)
      indent; println
      // bb foreach printInstruction
      for ((ins, idx) <- bb.toList.zipWithIndex) {
        print(idx)
        print("|\t")
        val msgConsumedTypes = ins.consumedTypes
        val msgProducedTypes = ins match {
          case cm : opcodes.CALL_METHOD => toTypeKind(cm.method.tpe.resultType) 
          case _ => ins.producedTypes
        }
        println(ins.toString())
        println("\t\t, consumed types: " + msgConsumedTypes)
        println("\t\t, produced types: " + msgProducedTypes)
      }
      undent; println
    }

    override def printMethod(m: IMethod) {
      println("/*-------------------------------------*/")
      print("def "); print(m.symbol.name);
      print("("); printList(printParam)(m.params, ", "); print(")");
      print(": "); print(m.symbol.info.resultType)

      if (!m.isDeferred) {
        println(" {")
        val localsWithTypes = m.locals map {l => (l, l.kind)}
        println("locals: " + localsWithTypes.mkString("[", ", ", "]"))
        println("startBlock: " + m.code.startBlock)
        println("blocks: " + m.code.blocks.mkString("[", ",", "]"))
        println
        lin.linearize(m) foreach printBlock
        println("}")

        indent; println("Exception handlers: ")
        m.exh foreach printExceptionHandler

        undent; println
      } else
        println
    }

  }


}

object GenCSharp
{
	import java.io.Writer
}